// Copyright 1992, 1999, 2011 by Jon Dart.  All Rights Reserved.

#ifndef _HASH_H
#define _HASH_H

#include "chess.h"
#include "stats.h"

extern const hash_t rep_codes[3];

class HashEntry;

class PositionInfo
{
   public:

      // Contents of the flag field
      enum {
         TYPE_MASK = 0x3,
         AGE_MASK = 0xc,
         CHECK_MASK = 0x10,
         FULL_MASK = 0x20,
         FORCED_MASK = 0x40,
         FORCED2_MASK = 0x80
      };

      // Only the first 3 values are actually stored - Invalid indicates
      // a hash hit with inadequate depth; NoHit indicates failure to find
      // a hash match.
      enum ValueType { Valid, UpperBound, LowerBound, Invalid, NoHit };

      PositionInfo() {
      }

      PositionInfo(int value, int depth,
         ValueType type, int age, int flags = 0,
      Move bestMove = NullMove) {
         contents.value = value;
         contents.depth = depth;
         contents.flags = type | (age << 2) | flags;
         contents.start = StartSquare(bestMove);
         contents.dest = DestSquare(bestMove);
         contents.promotion = PromoteTo(bestMove);
      }

      int depth() const
      {
         return contents.depth;
      }

      int value() const
      {
         return (int)contents.value;
      }

      void setValue(int value) {
         contents.value = (int16)value;
      }

      ValueType type() const
      {
         return (ValueType)(contents.flags & TYPE_MASK);
      }

      int age() const
      {
         return (int)((contents.flags & AGE_MASK) >> 2);
      }

      void setAge(int age) {
         contents.flags &= ~AGE_MASK;
         contents.flags |= (age << 2);
      }

      int inCheck() const
      {
         return (int)((contents.flags & CHECK_MASK) != 0);
      }

      int forced() const
      {
         return (int)((contents.flags & FORCED_MASK) != 0);
      }

      int forced2() const
      {
         return (int)((contents.flags & FORCED2_MASK) != 0);
      }

      int isFull() const
      {
         return (int)((contents.flags & FULL_MASK) != 0);
      }

      void setFull(int b) {
         if (b) contents.flags |= FULL_MASK;
         else contents.flags &= ~FULL_MASK;
      }

      Move bestMove(const Board &b) const
      {
         if (contents.start == InvalidSquare)
            return NullMove;
         else {
            Move m = CreateMove(b,(Square)contents.start,(Square)contents.dest,
               (PieceType)contents.promotion);
            //if (!validMove(b,m)) m = NullMove;
            if (forced()) SetForced(m);
            if (forced2()) SetForced2(m);
            return m;
         }
      }

   protected:
      union
      {
         struct
         {
            int16 value;
            int16 depth;
            byte flags;
            signed char start, dest, promotion;
         } contents;
         uint64 val;
      };

};

class HashEntry : public PositionInfo
{

   // this class represents one entry in the hash table.
   public:

      HashEntry( hash_t hash, const int depth,
         int age,
         PositionInfo::ValueType type,
         int inCheck,
         int value,
         int full,
         Move best_move)
         : PositionInfo(value, depth, type, age,
         (full ? FULL_MASK : 0) |
         (inCheck ? CHECK_MASK : 0) |
         (IsForced(best_move) ? FORCED_MASK : 0) |
         (IsForced2(best_move) ? FORCED2_MASK : 0),
      best_move) {
         hc = val ^ hash;
      }

  /*
      // optimized constructor
      HashEntry( hash_t hash, const int depth,
         int age,
         PositionInfo::ValueType type,
         int inCheck,
         int value,
         int full,
         Move bestMove) {
        val = (value & 0xff) | (depth << 16) & 0xff00 | (
          (uint64)(type | (age << 2) | 
          (full ? FULL_MASK : 0) |
          (inCheck ? CHECK_MASK : 0) |
          (IsForced(bestMove) ? FORCED_MASK : 0) |
          (IsForced2(bestMove) ? FORCED2_MASK : 0) |
          (byte)StartSquare(bestMove) << 8 |
          (byte)DestSquare(bestMove) << 16 |
          (byte)PromoteTo(bestMove) << 24) << 32);
        ASSERT(contents.start != -1);
        hc = val ^ hash;
      }
  */

      hash_t hashCode() const
      {
         return hc;
      }

      hash_t getEffectiveHash() const
      {
         return val ^ hc;
      }

      void setEffectiveHash(hash_t hash) {
         hc = hash ^ val;
      }

   protected:
      hash_t hc;
};

unsigned long hash_code(const HashEntry &p);

struct DualHashEntry
{
   HashEntry replace;
   HashEntry depth;
};

extern void initHash(uint64 bytes);

extern void resizeHash(uint64 bytes);

extern void freeHash();

extern void clearHash();

// put info from the external permanent hash table into the
// in-memory hash table
extern void loadLearnInfo();

extern PositionInfo::ValueType searchHash(const Board &b,
hash_t hashCode,
int alpha, int beta, int ply,
int depth, int age,
PositionInfo &entry
);

extern void storeHash(const HashEntry &newPos, hash_t hashCode,
Statistics *stats = NULL);
#endif
